/**
 * Copyright 2009-2012 Bull S.A.S.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.ow2.util.ee.deploy.api.deployable;

import org.ow2.util.archive.api.IArchive;


/**
 * This interface is used to represent any deployment object that can be used within EasyBeans.<br>
 * The EJB/WEB/RAR/EAR, etc components should extend this interface.
 * @param <T> the type of the deployable.
 * @author Florent Benoit
  */
public interface IDeployable<T extends IDeployable<T>> {

    /**
     * Gets the Archive of this deployable object.
     * @return the archive for this deployable.
     */
    IArchive getArchive();

    /**
     * If the Deployable has been unpacked, return the initial archive.
     * @return the deployable before it is unpacked.
     */
    T getOriginalDeployable();

    /**
     * If the Deployable has been unpacked, return the unpacked deployable.
     * @return the unpacked deployable associated to this deployable.
     */
    T getUnpackedDeployable();

    /**
     * If the Deployable has been unpacked, sets the unpacked deployable.
     * @param unpackedDeployable the unpacked deployable associated to this deployable.
     */
    void setUnpackedDeployable(final T unpackedDeployable);

    /**
     * If the EAR has been unpacked, return the initial archive.
     * @param originalDeployable the deployable before it is unpacked.
     */
    void setOriginalDeployable(final T originalDeployable);

    /**
     * Adds a deployable info that can be used by a deployer.
     * @param extension the implementation of the extension to register.
     */
    void addExtension(final IDeployableInfo extension);

    /**
     * Gets the extension for the given class.
     * @param extensionClass the class to used as a key
     * @return the extension else null if not found.
     */
    IDeployableInfo getExtension(final Class<?  extends IDeployableInfo> extensionClass);

    /**
     * Gets a shorter name for this deployable.
     * @return a shorter name based on the URL.
     */
    String getShortName();

    /**
     * Gets the module name for this deployable.
     * This module name is a name based on the short name but without any extension.
     * @return the module name of this deployable
     */
    String getModuleName();
}
