/**
 * Copyright 2009-2012 Bull S.A.S.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.ow2.util.ee.deploy.api.deployable.metadata;

import org.ow2.util.ee.deploy.api.deployable.IDeployable;
import org.ow2.util.ee.deploy.api.deployer.DeployerException;

/**
 * Deployable metadata factory.
 * @author Gael Lalire
 */
public abstract class DeployableMetadataManager {

    /**
     * The RI class name.
     */
    private static final String REFERENCE_IMPLEMENTATION_CLASS_NAME =
        "org.ow2.util.ee.deploy.impl.deployable.metadata.DeployableMetadataManagerImpl";

    /**
     * The current instance by thread.
     */
    private static final ThreadLocal<DeployableMetadataManager> INSTANCE =
        new InheritableThreadLocal<DeployableMetadataManager>();

    /**
     * @return the latest set DeployableMetadataManager in current thread or the
     *         RI
     * @throws DeployerException if no ArchiveScannerFactory set and the RI can
     *         not be accessed with current thread classloader
     */
    public static DeployableMetadataManager getInstance() throws DeployerException {
        DeployableMetadataManager deployableMetadataManager = INSTANCE.get();
        if (deployableMetadataManager == null) {
            ClassLoader classLoader = Thread.currentThread().getContextClassLoader();
            try {
                Class<? extends DeployableMetadataManager> clazz = classLoader.loadClass(REFERENCE_IMPLEMENTATION_CLASS_NAME)
                        .asSubclass(DeployableMetadataManager.class);
                deployableMetadataManager = clazz.newInstance();
                INSTANCE.set(deployableMetadataManager);
            } catch (Exception e) {
                throw new DeployerException("Unable to create RI", e);
            }
        }
        return deployableMetadataManager;
    }

    /**
     * Set the DeployableMetadataManager to use in current thread.
     * @param deployableMetadataManager the deployableMetadataManager
     */
    public static void setInstance(final DeployableMetadataManager deployableMetadataManager) {
        INSTANCE.set(deployableMetadataManager);
    }

    /**
     * Create metadata from a deployable with a registred factory.
     * @param deployable the deployable
     * @param deployableMetadataClass the class of metadata
     * @param <T> the deployable type
     * @param <Z> the metadata type
     * @return new metadata
     * @throws DeployableMetadataException if metadata can not be created
     */
    public abstract <T extends IDeployable<T>, Z extends IDeployableMetadata<T>> Z createDeployableMetadata(T deployable,
            Class<Z> deployableMetadataClass) throws DeployableMetadataException;

    /**
     * Add a factory.
     * @param clazz the supported deployable interface
     * @param deployableMetadataFactory the metadata factory
     * @param <T> the supported deployable class
     * @param <Z> the deployable metadata class
     * @return true if your factory is assigned to deployable type
     */
    public abstract <T extends IDeployable<T>, Z extends IDeployableMetadata<T>> boolean addDeployableMetadataFactory(
            Class<T> clazz, IDeployableMetadataFactory<T, Z> deployableMetadataFactory);

    /**
     * Remove a factory.
     * @param clazz the supported deployable interface
     * @param <T> the supported deployable class
     * @param <Z> the deployable metadata class
     * @param deployableMetadataFactory the metadata factory
     */
    public abstract <T extends IDeployable<T>, Z extends IDeployableMetadata<T>> void removeDeployableMetadataFactory(
            Class<T> clazz, IDeployableMetadataFactory<T, Z> deployableMetadataFactory);

}
