/**
 * Copyright 2009-2012 Bull S.A.S.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.ow2.util.ee.deploy.api.deployer;

import org.ow2.util.ee.deploy.api.deployable.IDeployable;
import org.ow2.util.ee.deploy.api.report.IDeploymentReport;

import java.util.List;

/**
 * Interface allowing to deploy/undeploy the {@link IDeployable} objects.
 * @author Florent Benoit
 */
public interface IDeployer<T extends IDeployable<T>> {

    /**
     * Deploy a deployable. It can be an EJB jar, EAR, WAR, etc.
     * @param deployable a given deployable
     * @throws DeployerException if the deployment is not done.
     * @throws UnsupportedDeployerException No deployer found for deployable.
     */
    void deploy(final IDeployable<T> deployable) throws DeployerException, UnsupportedDeployerException;

    /**
     * Undeploy the given deployable. It can be an EJB jar, EAR, WAR, etc.
     * @param deployable a given deployable to undeploy
     * @throws DeployerException if the undeploy operation fails.
     */
    void undeploy(final IDeployable<T> deployable) throws DeployerException;

    /**
     * Checks if the given deployable is deployed or not.
     * @param deployable test if a given deployable is already deployed.
     * @return true if the given deployable is deployed.
     * @throws DeployerException if the isDeployed operation fails.
     */
    boolean isDeployed(final IDeployable<T> deployable) throws DeployerException;


    /**
     * Checks if the given deployable is supported by the Deployer.
     * @param deployable the deployable to be checked
     * @return true if it is supported, else false.
     */
    boolean supports(final IDeployable<?> deployable);

    /**
     * Deploy deployables. It can be an EJB jar, EAR, WAR, etc. List of same type deployable object
     *
     * @param deployables List of deployables to deploy
     * @return the list of {@link IDeploymentReport}
     */
    List<IDeploymentReport> deploy(final List<IDeployable<T>> deployables);

    /**
     * Undeploy the list of deployable. It can be an EJB jar, EAR, WAR, etc.
     * @param deployables List of deployable to undeploy
     * @return the list of {@link IDeploymentReport}
     */
    List<IDeploymentReport> undeploy(final List<IDeployable<T>> deployables);

    /**
     * Checks if this deployable list is supported by the Deployer.
     * @param deployables the list of deployable to be checked
     * @return true if it is supported, else false.
     */
    boolean supports(final List<IDeployable<?>> deployables);
}
