/**
 * Copyright 2009-2012 Bull S.A.S.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.ow2.util.ee.deploy.impl.deployable.factory;

import java.util.Collection;
import java.util.HashSet;

import org.ow2.util.archive.api.IFileArchive;
import org.ow2.util.log.Log;
import org.ow2.util.log.LogFactory;
import org.ow2.util.ee.deploy.api.deployable.FileDeployable;
import org.ow2.util.ee.deploy.api.deployable.factory.FileDeployableException;
import org.ow2.util.ee.deploy.api.deployable.factory.IFileDeployableFactory;
import org.ow2.util.ee.deploy.api.deployable.factory.IFileDeployableFactoryManager;

/**
 * A IFileDeployableFactoryManager that owns a collection of managed factories
 * and tries to build a deployable with them one by one.
 * @author mleduque
 */
public class DefaultFileDeployableFactoryManager implements IFileDeployableFactoryManager {
    /**
     * Logger.
     */
    private static Log logger = LogFactory.getLog(DefaultFileDeployableFactoryManager.class);

    /**
     * FileDeployableFactory objects managed by the manager.
     */
    private Collection<IFileDeployableFactory> fileDeployableFactories = new HashSet<IFileDeployableFactory>();

    /**
     * {@inheritDoc}
     */
    public FileDeployable<?, ?> getFileDeployable(final IFileArchive fileArchive) throws FileDeployableException {
        FileDeployable<?, ?> fileDeployable = null;
        for (IFileDeployableFactory factory : fileDeployableFactories) {
            try {
                fileDeployable = factory.getFileDeployable(fileArchive);
                return fileDeployable;
            } catch (FileDeployableException e) {
                logger.warn("factory {0} doesn''t support archive {1}", factory, fileArchive, e);
            }
        }
        /* No factory in the manager supports the archive type */
        throw new FileDeployableException();
    }

    /**
     * Adds a factory to the list of managed factories.
     * @param factory the added factory
     */
    public void addFileDeployableFactory(final IFileDeployableFactory factory) {
        fileDeployableFactories.add(factory);
    }

    /**
     * Removes a factory from the list of managed factories.
     * @param factory the removed factory.
     */
    public void removeFileDeployableFactory(final IFileDeployableFactory factory) {
        fileDeployableFactories.remove(factory);
    }
}
