/**
 * Copyright 2009-2012 Bull S.A.S.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.ow2.util.ee.deploy.impl.deployable.factory;

import java.io.IOException;
import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.util.HashMap;
import java.util.Map;

import javax.xml.parsers.ParserConfigurationException;

import org.ow2.util.archive.api.IFileArchive;
import org.ow2.util.log.Log;
import org.ow2.util.log.LogFactory;
import org.ow2.util.ee.deploy.api.deployable.FileDeployable;
import org.ow2.util.ee.deploy.api.deployable.factory.FileDeployableException;
import org.ow2.util.ee.deploy.api.deployable.factory.XmlFileDeployableFactory;
import org.ow2.util.ee.deploy.impl.deployable.factory.utils.XmlHelper;
import org.xml.sax.SAXException;

/**
 * Implementation for the XmlFileDeployableFactory interface.
 * @author Mickaël LEDUQUE
 */
public class XmlFileDeployableFactoryImpl implements XmlFileDeployableFactory {

    /**
     * The logger.
     */
    private Log logger = LogFactory.getLog(XmlFileDeployableFactoryImpl.class);

    /**
     * The XmlHelper used to identify XML files.
     */
    private XmlHelper xmlHelper = new XmlHelper();

    /**
     * Default constructor.
     * @throws ParserConfigurationException when the XmlHelper initilization
     *         fails.
     * @throws SAXException when the XmlHelper initilization fails.
     */
    public XmlFileDeployableFactoryImpl() throws ParserConfigurationException, SAXException {
        /* Needed because xmlHelper initialization can throw exceptions */
    }

    /**
     * XML namespace to FileDeployable association.
     */
    private Map<String, Class<? extends FileDeployable<?, ?>>> xmlNSToDeployableMap
        = new HashMap<String, Class<? extends FileDeployable<?, ?>>>();

    /**
     * Returns the file deployable constructed from the IFileArchive using the
     * given namespace.
     * @param fileArchive the archive that must be converted.
     * @param nameSpace the namespace to be used.
     * @return the FileDeployable.
     * @throws FileDeployableException when the Deployable could not be
     *         constructed.
     */
    private FileDeployable<?, ?> getXmlFileDeployable(final IFileArchive fileArchive, final String nameSpace)
            throws FileDeployableException {
        Class<? extends FileDeployable<?, ?>> clazz = xmlNSToDeployableMap.get(nameSpace);
        if (clazz == null) {
            logger.debug("The XML namespace {0} found in the archive ''{0}'' is not handled by a factory", nameSpace,
                    fileArchive);
            return null;
        }
        Constructor<? extends FileDeployable<?, ?>> constructor = null;
        try {
            constructor = clazz.getDeclaredConstructor(IFileArchive.class);
        } catch (NoSuchMethodException e) {
            logger.error("No {0}(IArchive archive) constructor; aborting deployable generation", clazz.getName());
            throw new FileDeployableException(e);
        }
        try {
            return constructor.newInstance(fileArchive);
        } catch (InvocationTargetException e) {
            /*
             * Catch the InvocationTargetException because the nested exception
             * is more meaningful.
             */
            Throwable nested = e.getCause();
            throw new FileDeployableException(nested);
        } catch (Exception e) {
            logger.error("Caught error on constructor invokation : {0}", e);
            throw new FileDeployableException();
        }
    }

    /**
     * {@inheritDoc}
     */
    public FileDeployable<?, ?> getFileDeployable(final IFileArchive fileArchive) throws FileDeployableException {
        String namespace = null;
        try {
            namespace = xmlHelper.getRootElementNameSpace(fileArchive.getFile());
        } catch (IOException e) {
            logger.error("IO Exception while parsing file {0}", fileArchive.getFile(), e);
            throw new FileDeployableException(e);
        } catch (SAXException e) {
            logger.error("Exception while parsing file {0} : this is not a XML file", fileArchive.getFile(), e);
            throw new FileDeployableException(e);
        }

        return getXmlFileDeployable(fileArchive, namespace);
    }

    /**
     * {@inheritDoc}
     */
    public void registerFileDeployable(final Class<? extends FileDeployable<?, ?>> clazz, final String nameSpace) {
        xmlNSToDeployableMap.put(nameSpace, clazz);
    }

    /**
     * {@inheritDoc}
     */
    public void unregisterFileDeployable(final String nameSpace) {
        xmlNSToDeployableMap.remove(nameSpace);

    }

}
