/**
 * Copyright 2009-2012 Bull S.A.S.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.ow2.util.ee.deploy.impl.deployable.factory.utils;

import java.io.File;
import java.io.IOException;

import javax.xml.parsers.ParserConfigurationException;
import javax.xml.parsers.SAXParser;
import javax.xml.parsers.SAXParserFactory;

import org.xml.sax.SAXException;
import org.xml.sax.helpers.DefaultHandler;

public class XmlHelper {

    /**
     * The parser that will be used for validation with namespaces.
     */
    private SAXParser namespaceAwareParser = null;

    /**
     * The parser that will be used for parsing when namespaces are not
     * considered.
     */
    private SAXParser namespaceUnawareParser = null;

    /**
     * Default constructor for XmlHelperImpl.
     *
     * @throws ParserConfigurationException in case of SAX misconfiguration.
     * @throws SAXException is SAX errors occured.
     */
    public XmlHelper() throws ParserConfigurationException, SAXException {
        SAXParserFactory factory = SAXParserFactory.newInstance();;
        namespaceUnawareParser = factory.newSAXParser();
        factory.setNamespaceAware(true);
        namespaceAwareParser = factory.newSAXParser();
    }

    /**
     * {@inheritDoc}
     */
    public synchronized boolean isWellFormedXml(final File file) throws IllegalArgumentException, IOException {
        try {
            namespaceUnawareParser.parse(file, new DefaultHandler());
            return true;
        } catch (SAXException e) {
            return false;
        }
    }

    /**
     * {@inheritDoc}
     */
    public synchronized boolean isLikelyXml(final File file) throws IllegalArgumentException, IOException {
        try {
            namespaceUnawareParser.parse(file, new CheckFirstElementHandler());
            return false;
        } catch (GetOutException e) {
            return true;
        } catch (SAXException e) {
            return false;
        }

    }

    /**
     * {@inheritDoc}
     */
    public synchronized String getRootElementNameSpace(final File file) throws IOException, SAXException {
        try {
            namespaceAwareParser.parse(file, new CheckFirstElementHandler());
            return null; // else the compiler is unhappy
        } catch (GetOutException e) {
            return e.getNameSpace();
        }
    }

    /**
     * Exception used to cancel the parsing once the first element
     * of an XML file has been read.
     * @author mleduque
     */
    static class GetOutException extends SAXException {

        private static final long serialVersionUID = 27568468415776L;

        /**
         * The namespace of the first element of the file.
         */
        private String nameSpace = null;

        /**
         * the first element name.
         */
        private String elementName = null;

        /**
         *
         * @param nameSpace the name space.
         * @param elementName the element name.
         */
        public GetOutException(final String nameSpace, final String localName) {
            this.nameSpace = nameSpace;
            this.elementName = localName;
        }

        /**
         * Sets the namespace.
         * @param nameSpace the new namespace.
         */
        public void setNameSpace(final String nameSpace) {
            this.nameSpace = nameSpace;
        }

        /**
         * Gets the namespace.
         * @return the namespace.
         */
        public String getNameSpace() {
            return nameSpace;
        }

        /**
         * Sets the element name.
         * @param elementName the new element name.
         */
        public void setElementName(final String localName) {
            this.elementName = localName;
        }

        /**
         * Gets the element name.
         * @return the element name.
         */
        public String getElementName() {
            return this.elementName;
        }
    }

    /**
     * SAX handler for the xml helper.
     * @author mleduque
     */
    static class CheckFirstElementHandler extends DefaultHandler {

        /**
         * {@inheritDoc}
         */
        @Override
        public void startElement(final String uri, final String localName, final String qName,
                final org.xml.sax.Attributes attributes) throws SAXException {
                /* We've found the root element
                 * Abort the parsing. */
                throw new GetOutException(uri, localName);
        }

    }
}
