/**
 * Copyright 2009-2012 Bull S.A.S.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.ow2.util.ee.deploy.impl.deployable.metadata;

import java.util.HashMap;
import java.util.Map;

import org.ow2.util.ee.deploy.api.deployable.IDeployable;
import org.ow2.util.ee.deploy.api.deployable.metadata.DeployableMetadataException;
import org.ow2.util.ee.deploy.api.deployable.metadata.DeployableMetadataManager;
import org.ow2.util.ee.deploy.api.deployable.metadata.IDeployableMetadata;
import org.ow2.util.ee.deploy.api.deployable.metadata.IDeployableMetadataFactory;

/**
 * DefaultDeployableMetadataFactory.
 * @author Gael Lalire
 */
public class DeployableMetadataManagerImpl extends DeployableMetadataManager {

    /**
     * Factory map.
     */
    private Map<Class<?>, Map<Class<?>, IDeployableMetadataFactory<?, ?>>> factoryMap;

    /**
     * Constructor.
     */
    public DeployableMetadataManagerImpl() {
        factoryMap = new HashMap<Class<?>, Map<Class<?>, IDeployableMetadataFactory<?, ?>>>();
    }

    /**
     * Create metadata from a deployable with a registred factory.
     * @param deployable the deployable
     * @param deployableMetadataClass the class of metadata
     * @param <T> the deployable type
     * @param <Z> the metadata type
     * @return new metadata
     * @throws DeployableMetadataException if metadata can not be created
     */
    @SuppressWarnings("unchecked")
    @Override
    public <T extends IDeployable<T>, Z extends IDeployableMetadata<T>> Z createDeployableMetadata(final T deployable,
            final Class<Z> deployableMetadataClass) throws DeployableMetadataException {
        Map<Class<?>, IDeployableMetadataFactory<?, ?>> map = factoryMap.get(deployableMetadataClass);
        IDeployableMetadataFactory<T, Z> factory = null;
        for (Class<?> clazz : deployable.getClass().getInterfaces()) {
            IDeployableMetadataFactory<?, ?> factoryTmp = map.get(clazz);
            if (factoryTmp != null) {
                factory = (IDeployableMetadataFactory<T, Z>) factoryTmp;
                break;
            }
        }
        if (factory == null) {
            throw new DeployableMetadataException("No factory found for the type '" + deployableMetadataClass + "'.");
        }
        return factory.createDeployableMetadata(deployable);
    }

    /**
     * Add a factory.
     * @param clazz the supported deployable interface
     * @param deployableMetadataFactory the metadata factory
     * @param <T> the supported deployable class
     * @param <Z> the deployable metadata class
     * @return true if your factory is assigned to deployable type
     */
    @Override
    public <T extends IDeployable<T>, Z extends IDeployableMetadata<T>> boolean addDeployableMetadataFactory(
            final Class<T> clazz, final IDeployableMetadataFactory<T, Z> deployableMetadataFactory) {
        Class<?> deployableMetadataClass = deployableMetadataFactory.getDeployableMetadataClass();
        Map<Class<?>, IDeployableMetadataFactory<?, ?>> map = factoryMap.get(deployableMetadataClass);
        if (map == null) {
            map = new HashMap<Class<?>, IDeployableMetadataFactory<?, ?>>();
            factoryMap.put(deployableMetadataClass, map);
        }
        if (!map.containsKey(deployableMetadataClass)) {
            map.put(clazz, deployableMetadataFactory);
            return true;
        }
        return false;
    }

    /**
     * Remove a factory.
     * @param clazz the supported deployable interface
     * @param <T> the supported deployable class
     * @param <Z> the deployable metadata class
     * @param deployableMetadataFactory the metadata factory
     */
    @Override
    public <T extends IDeployable<T>, Z extends IDeployableMetadata<T>> void removeDeployableMetadataFactory(
            final Class<T> clazz, final IDeployableMetadataFactory<T, Z> deployableMetadataFactory) {
        Class<?> deployableMetadataClass = deployableMetadataFactory.getDeployableMetadataClass();
        Map<Class<?>, IDeployableMetadataFactory<?, ?>> map = factoryMap.get(deployableMetadataClass);
        if (map == null) {
            return;
        }
        IDeployableMetadataFactory<?, ?> factory = map.get(clazz);
        if (factory.equals(deployableMetadataFactory)) {
            factoryMap.remove(deployableMetadataClass);
        }
    }

}
