/**
 * Copyright 2009-2012 Bull S.A.S.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.ow2.util.ee.deploy.impl.osgi;

import org.osgi.framework.BundleActivator;
import org.osgi.framework.BundleContext;
import org.osgi.framework.InvalidSyntaxException;
import org.osgi.framework.ServiceEvent;
import org.osgi.framework.ServiceListener;
import org.osgi.framework.ServiceReference;
import org.ow2.util.ee.deploy.api.deployer.IDeployerManagerReportCallback;
import org.ow2.util.ee.deploy.api.helper.IDeployableHelper;
import org.ow2.util.archive.api.IArchiveManager;
import org.ow2.util.archive.impl.ArchiveManager;
import org.ow2.util.ee.deploy.api.deployer.IDeployer;
import org.ow2.util.ee.deploy.api.deployer.IDeployerManager;
import org.ow2.util.ee.deploy.api.deployer.IDeployerManagerCallback;
import org.ow2.util.ee.deploy.impl.deployer.DeployerManager;
import org.ow2.util.ee.deploy.impl.helper.DeployableAnalyserTracker;
import org.ow2.util.ee.deploy.impl.helper.DeployableHelperComponent;

/**
 * Activator which registers DeployerManager, ArchiveManager and OSGi Deployer
 * services.
 * @author Francois Fornaciari
 */
public class Activator implements BundleActivator, ServiceListener {

    /**
     * The bundle context.
     */
    private BundleContext bundleContext = null;

    /**
     * The deployer manager.
     */
    private DeployerManager deployerManager = null;

    /**
     * The DeployableAnalyser tracker
     */
    private DeployableAnalyserTracker deployableAnalyserTracker = null;

    /**
     * Called when this bundle is started so the Framework can perform the
     * bundle-specific activities necessary to start this bundle. This method
     * can be used to register services or to allocate any resources that this
     * bundle needs.
     * @param bundleContext The execution context of the bundle being started.
     * @throws Exception If the start fails
     */
    public void start(final BundleContext bundleContext) throws Exception {
        this.bundleContext = bundleContext;

        // Create the OSGi deployer
        OSGiDeployer osgiDeployer = new OSGiDeployer(bundleContext);

        // Create a new deployer manager
        this.deployerManager = new DeployerManager();

        // Register the deployer on the deployer manager
        this.deployerManager.register(osgiDeployer);

        // Create a new DeployableHelperComponent
        this.deployableAnalyserTracker = new DeployableAnalyserTracker(bundleContext);
        this.deployableAnalyserTracker.start();  //start the DeployableAnalyser tracker
        DeployableHelperComponent deployableHelperComponent = new DeployableHelperComponent(this.deployableAnalyserTracker);

        // Register the deployer manager and the deployer and the DeployerHelperComponent as an OSGi service
        bundleContext.registerService(IDeployerManager.class.getName(), this.deployerManager, null);
        bundleContext.registerService(IDeployer.class.getName(), osgiDeployer, null);
        bundleContext.registerService(IDeployableHelper.class.getName(), deployableHelperComponent, null);

        // Add also the archive manager
        ArchiveManager archiveManager = ArchiveManager.getInstance();
        bundleContext.registerService(IArchiveManager.class.getName(), archiveManager, null);

        // Add existing DeployerManager callbacks registered as OSGi services
        addCallbacks();
        
        // Handle callback registrations/unregistrations
        bundleContext.addServiceListener(this, "(objectClass=" + IDeployerManagerCallback.class.getName() + ")");
        bundleContext.addServiceListener(this, "(objectClass=" + IDeployerManagerReportCallback.class.getName() + ")");
    }

    /**
     * Called when this bundle is stopped so the Framework can perform the
     * bundle-specific activities necessary to stop the bundle.
     * @param bc The execution context of the bundle being stopped.
     */
    public void stop(final BundleContext bc) {
        //stop the DeployableAnalyser tracker
        this.deployableAnalyserTracker.stop();
    }

    /**
     * {@inheritDoc}
     */
    public void serviceChanged(ServiceEvent event) {
        ServiceReference serviceRef = event.getServiceReference();
        Object object = this.bundleContext.getService(serviceRef);
        switch (event.getType()) {
            case ServiceEvent.REGISTERED:
                if (object instanceof IDeployerManagerCallback) {
                    this.deployerManager.addCallback(IDeployerManagerCallback.class.cast(object));
                } else if (object instanceof IDeployerManagerReportCallback) {
                    this.deployerManager.addCallback(IDeployerManagerReportCallback.class.cast(object));
                }
                break;
            case ServiceEvent.UNREGISTERING:
                if (object instanceof IDeployerManagerCallback) {
                    this.deployerManager.removeCallback(IDeployerManagerCallback.class.cast(object));
                } else if (object instanceof IDeployerManagerReportCallback) {
                    this.deployerManager.removeCallback(IDeployerManagerReportCallback.class.cast(object));
                }
                break;
            default:
                break;
        }
    }

    /**
     * Add a new {@IDeplyerManagerCallback} registered as an OSGi service.
     * @throws InvalidSyntaxException If the service reference cannot be retrieved
     */
    private void addCallbacks() throws InvalidSyntaxException {
        ServiceReference[] serviceReferences = this.bundleContext.getServiceReferences(
                IDeployerManagerCallback.class.getName(), null);
        if (serviceReferences != null) {
            for (ServiceReference serviceReference : serviceReferences) {
                IDeployerManagerCallback callback = (IDeployerManagerCallback) this.bundleContext.getService(serviceReference);
                this.deployerManager.addCallback(callback);
            }
        }
        serviceReferences = this.bundleContext.getServiceReferences(IDeployerManagerReportCallback.class.getName(), null);
        if (serviceReferences != null) {
            for (ServiceReference serviceReference : serviceReferences) {
                IDeployerManagerReportCallback callback = (IDeployerManagerReportCallback) this.bundleContext.getService(serviceReference);
                this.deployerManager.addCallback(callback);
            }
        }
    }

}
