/**
 * Copyright 2009-2012 Bull S.A.S.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.ow2.util.ee.deploy.impl.osgi;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.osgi.framework.Bundle;
import org.osgi.framework.BundleContext;
import org.osgi.framework.BundleException;
import org.osgi.framework.startlevel.BundleStartLevel;
import org.ow2.util.archive.api.IArchiveMetadata;
import org.ow2.util.ee.deploy.api.deployable.IDeployable;
import org.ow2.util.ee.deploy.api.deployable.OSGiDeployable;
import org.ow2.util.ee.deploy.api.deployer.DeployerException;
import org.ow2.util.ee.deploy.api.report.IDeploymentReport;
import org.ow2.util.ee.deploy.impl.deployer.AbsDeployer;
import org.ow2.util.ee.deploy.impl.report.DeploymentReport;
import org.ow2.util.log.Log;
import org.ow2.util.log.LogFactory;

/**
 * OSGi deployer that is able to deploy an OSGi Deployable
 * @author Florent BENOIT
 * @author Jeremy Cazaux (support to the deployment of a list of deployables)
 */
public class OSGiDeployer extends AbsDeployer<OSGiDeployable> {

    /**
     * Logger.
     */
    private Log logger = LogFactory.getLog(OSGiDeployer.class);

    /**
     * Bundle Context to use for the deployment.
     */
    private BundleContext bundleContext = null;

    /**
     * List of Bundles that have been installed.
     */
    private Map<String, Bundle> installedBundles = null;

    /**
     * Bundle symbolic name property
      */
    public static String BUNDLE_SYMBOLIC_NAME = "Bundle-SymbolicName";

    /**
     * Bundle version property
     */
    public static String BUNDLE_VERSION = "Bundle-Version";

    /**
     * Default constructor.
     * @param bundleContext OSGi bundle context
     */
    public OSGiDeployer(final BundleContext bundleContext) {
        this.bundleContext = bundleContext;
        this.installedBundles = new HashMap<String, Bundle>();
    }

    /**
     * Undeploy the given Bundle.
     * @param osgiDeployable the deployable to remove.
     * @throws org.ow2.util.ee.deploy.api.deployer.DeployerException if the bundle is not undeployed.
     */
    public void doUndeploy(final IDeployable<OSGiDeployable> osgiDeployable) throws DeployerException {
        logger.debug("Undeploying {0}", osgiDeployable);

        //stop the bundle
        Bundle bundle = stopBundle(osgiDeployable);

        if (bundle != null) {
            //uninstall the bundle
            uninstallBundle(osgiDeployable, bundle);
        }
    }


    /**
     * Deploy the given Bundle.
     * @param deployable the deployable to add.
     * @throws org.ow2.util.ee.deploy.api.deployer.DeployerException if the Bundle is not deployed.
     */
    public void doDeploy(final IDeployable<OSGiDeployable> deployable) throws DeployerException {
        logger.debug("Deploying {0}", deployable);

        // Cast the deployable only once
        OSGiDeployable osgiDeployable = OSGiDeployable.class.cast(deployable);

        // First, install the bundle
        Bundle bundle = installBundle(osgiDeployable);

        if (bundle != null) {
            // And then start it
            startBundle(osgiDeployable, bundle);

            // Add it as installed
            addInstalledBundle(osgiDeployable, bundle);
        }
    }

    /**
     * Checks if the given deployable is supported by the Deployer.
     * @param deployable the deployable to be checked
     * @return true if it is supported, else false.
     */
    public boolean supports(final IDeployable<?> deployable) {
        return OSGiDeployable.class.isInstance(deployable);
    }

    /**
     * Check if the given deployable is deployable or not.
     * @param deployable the deployable to check.
     * @throws org.ow2.util.ee.deploy.api.deployer.DeployerException if the deployable is not supported.
     */
    protected void check(final IDeployable<?> deployable) throws DeployerException {
        if (bundleContext == null) {
            throw new DeployerException("The deployment of the deployable'" + deployable
                    + "' is not available as there is no bundle context set on this deployer.");
        }

        super.check(deployable);
    }

    /**
     * Deploy the list of given Bundle.
     * @param osgiDeployables the list of OSGi deployable to add.
     * @return the list of {@link org.ow2.util.ee.deploy.api.report.IDeploymentReport}
     */
    @Override
    public List<IDeploymentReport> doDeploy(List<IDeployable<OSGiDeployable>> osgiDeployables) {

        //list of deployment reports
        List<IDeploymentReport> deploymentReports = new ArrayList<IDeploymentReport>();

        Map<OSGiDeployable, Bundle> osgiDeployableBundleMap = new HashMap<OSGiDeployable, Bundle>();

        // first, install all bundles
        for (IDeployable<OSGiDeployable> deployable : osgiDeployables) {
            this.logger.debug("Deploying {0}", deployable);

            // Cast the deployable only once
            OSGiDeployable osgiDeployable = OSGiDeployable.class.cast(deployable);

            //install the bundle
            Bundle bundle;
            try {
                bundle = installBundle(osgiDeployable);
                if (bundle != null) {
                    //Keep the link OSGiDeployable -> Bundle
                    osgiDeployableBundleMap.put(osgiDeployable, bundle);
                }
            } catch (DeployerException e) {
                logger.error("Bundle {0} cannot be installed", osgiDeployable, e);
                DeploymentReport deploymentReport = new DeploymentReport();
                deploymentReport.setDeployable(osgiDeployable);
                deploymentReport.setDeploymentOk(false);
                deploymentReport.setException(e);
                deploymentReports.add(deploymentReport);
            }
        }

        //And then start all bundles
        for (Map.Entry<OSGiDeployable, Bundle> entry : osgiDeployableBundleMap.entrySet()) {

            OSGiDeployable osgiDeployable = entry.getKey();
            Bundle bundle = entry.getValue();

            //start the bundle
            startBundle(osgiDeployable, bundle);

            // Add it as installed
            try {
                addInstalledBundle(osgiDeployable, bundle);
                DeploymentReport deploymentReport = new DeploymentReport();
                deploymentReport.setDeployable(osgiDeployable);
                deploymentReport.setDeploymentOk(true);
                deploymentReports.add(deploymentReport);
            } catch (DeployerException e) {
                logger.error("Cannot add {0} to the list of installed bundle", osgiDeployable, e);
                DeploymentReport deploymentReport = new DeploymentReport();
                deploymentReport.setDeployable(osgiDeployable);
                deploymentReport.setDeploymentOk(false);
                deploymentReport.setException(e);
                deploymentReports.add(deploymentReport);
            }

        }
        return deploymentReports;
    }

    /**
     * Undeploy a list of Bundle without dependencies problem.
     * @param osgiDeployables The list of deployable to remove.
     * @return the list of {@link org.ow2.util.ee.deploy.api.report.IDeploymentReport}
     */
    @Override
    public List<IDeploymentReport> doUndeploy(final List<IDeployable<OSGiDeployable>> osgiDeployables) {

        //list of deployment reports
        List<IDeploymentReport> deploymentReports = new ArrayList<IDeploymentReport>();

        Map<IDeployable<OSGiDeployable>, Bundle> osgiDeployableBundleMap = new HashMap<IDeployable<OSGiDeployable>, Bundle>();

        for (IDeployable<OSGiDeployable> osgiDeployable: osgiDeployables) {

            this.logger.debug("Undeploying {0}", osgiDeployable);

            Bundle bundle;
            try {
                bundle = stopBundle(osgiDeployable);
                //keep the link deployable -> bundle if the bundle is not already uninstalled
                if (bundle != null) {
                    osgiDeployableBundleMap.put(osgiDeployable, bundle);
                }
            } catch (DeployerException e) {
                logger.error("Cannot stop bundle {0}", osgiDeployable, e);
                DeploymentReport deploymentReport = new DeploymentReport();
                deploymentReport.setDeployable(osgiDeployable);
                deploymentReport.setDeploymentOk(false);
                deploymentReport.setException(e);
                deploymentReports.add(deploymentReport);
            }
        }

        //And then uninstall all installed bundles
        for (Map.Entry<IDeployable<OSGiDeployable>, Bundle> entry : osgiDeployableBundleMap.entrySet()) {

            IDeployable<OSGiDeployable> osgiDeployable = entry.getKey();

            try {
                uninstallBundle(osgiDeployable, entry.getValue());
                DeploymentReport deploymentReport = new DeploymentReport();
                deploymentReport.setDeployable(osgiDeployable);
                deploymentReport.setDeploymentOk(true);
                deploymentReports.add(deploymentReport);
            } catch (DeployerException e) {
                logger.error("Cannot uninstall bundle {0}", osgiDeployable, e);
                DeploymentReport deploymentReport = new DeploymentReport();
                deploymentReport.setDeployable(osgiDeployable);
                deploymentReport.setDeploymentOk(false);
                deploymentReport.setException(e);
                deploymentReports.add(deploymentReport);
            }
        }
        return deploymentReports;
    }


    /**
     * Install an OSGi deployable
     * @param osgiDeployable The OSGi deployable to install
     * @return the installed bundle if the OSGi deployable has been installed. Return null if the bundle is already
     * installed.
     * @throws org.ow2.util.ee.deploy.api.deployer.DeployerException
     */
    protected Bundle installBundle(OSGiDeployable osgiDeployable) throws DeployerException {
        Bundle osgiBundle;
        try {
            String url = getURL(osgiDeployable).toExternalForm();
            if (osgiDeployable.isReference()) {
                url = "reference:".concat(url);
            }

            //install the bundle
            osgiBundle = this.bundleContext.installBundle(url);
            osgiBundle.adapt(BundleStartLevel.class).setStartLevel(osgiDeployable.getStartLevel());
            

        } catch (BundleException e) {
            // Check the bundle exception type
            if (e.getType() == BundleException.DUPLICATE_BUNDLE_ERROR) {
                logger.warn("Cannot deploy a bundle that is already installed " + osgiDeployable + ". ",  e.getMessage());
                return null;
            } else {
                throw new DeployerException("Cannot install the bundle '" + osgiDeployable + "'.", e);
            }
        }

        return osgiBundle;
    }

    /**
     * @param osgiDeployable The OSGi deployable
     * @param bundle Its associated Bundle to start
     */
    protected void startBundle(OSGiDeployable osgiDeployable, Bundle bundle) {

        if (osgiDeployable.isStart()) {

            try {
                // Compute the options
                // Always respect the bundle's activation policy
                int options = Bundle.START_ACTIVATION_POLICY;

                // Enable transient bundle if required
                if (osgiDeployable.isStartTransient()) {
                    options |= Bundle.START_TRANSIENT;
                }

                // Start the bundle with the options
                bundle.start(options);
            } catch (BundleException e) {
                logger.error("Cannot start the bundle ''{0}''.", osgiDeployable, e);
            }
        }
    }

    /**
     * @param osgiDeployable The OSGi deployable
     * @param bundle Its associated Bundle
     * @throws org.ow2.util.ee.deploy.api.deployer.DeployerException
     */
    private void addInstalledBundle(OSGiDeployable osgiDeployable, Bundle bundle) throws DeployerException {
        this.installedBundles.put(getURL(osgiDeployable).toExternalForm(), bundle);
    }

    /**
     *
     * @param osgiDeployable The OSGi deployable
     * @return its associated bundle if the OSGi deployable has been stop
     * @throws org.ow2.util.ee.deploy.api.deployer.DeployerException is the deployable is already uninstalled
     */
    protected Bundle stopBundle(IDeployable<OSGiDeployable> osgiDeployable) throws DeployerException {

        Bundle bundle;

        // get the bundle Object
        if (isDeployed(osgiDeployable)) {
            bundle = installedBundles.get(getURL(osgiDeployable).toExternalForm());
        } else {
            //TODO the deployment system should be redesigned in order to standardize the management of deployables cache.

            //if the deployable is not yet marked as deployed, maybe the OSGi bundle is still available in the OSGi cache
            bundle = getBundle(osgiDeployable);
            if (bundle != null) {
                this.installedBundles.put(getURL(osgiDeployable).toExternalForm(), bundle);
            }
        }

        // Check found
        if (bundle == null) {
            throw new DeployerException("No bundle found for the '" + osgiDeployable + "'.");
        }

        // Do not stop or uninstall if already uninstall/removed
        if (Bundle.UNINSTALLED != bundle.getState()) {
            // Stop the bundle
            try {
                bundle.stop();
            } catch (BundleException e) {
                throw new DeployerException("Cannot stop the bundle '" + osgiDeployable + "'.", e);
            }

            return bundle;

        } else {

            // Remove it of installed bundle
            this.installedBundles.remove(getURL(osgiDeployable).toExternalForm());

            return null;
        }
    }

    /**
     * Uninstalled a bundle
     * @param osgiDeployable The OSGi deployable
     * @param bundle Its associated bundle
     * @throws org.ow2.util.ee.deploy.api.deployer.DeployerException if the bundle cannot be uninstalled
     */
    protected void uninstallBundle(IDeployable<OSGiDeployable> osgiDeployable, Bundle bundle) throws DeployerException {

        // Uninstall the bundle
        try {
            bundle.uninstall();
        } catch (BundleException e) {
            throw new DeployerException("Cannot uninstall the bundle '" + osgiDeployable + "'.", e);
        }

        // Remove it of installed bundle
        this.installedBundles.remove(getURL(osgiDeployable).toExternalForm());
    }


    /**
     * @param osgiDeployable An OSGi deployable
     * @return the associated {@link Bundle} from the OSGi plateform
     */
    private Bundle getBundle(IDeployable<OSGiDeployable> osgiDeployable) {
        //get the bundle key
        IArchiveMetadata archiveMetadata = osgiDeployable.getArchive().getMetadata();
        String symbolicName = archiveMetadata.get(BUNDLE_SYMBOLIC_NAME);
        String version = archiveMetadata.get(BUNDLE_VERSION);

        //scan the OSGi platform in order to retrieve the right Bundle
        if (symbolicName != null && version != null) {
            List<Bundle> bundles = Arrays.asList(this.bundleContext.getBundles());
            for (Bundle bundle: bundles) {
                if (bundle.getSymbolicName().equals(symbolicName) && bundle.getVersion().toString().equals(version)) {
                    return bundle;
                }
            }
        }
        return null;
    }
}
