/**
 * Copyright 2009-2012 Bull S.A.S.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.ow2.util.ee.deploy.test;


import java.io.BufferedReader;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.Writer;
import java.net.URL;
import java.net.URLConnection;
import java.util.ArrayList;
import java.util.List;
import java.util.jar.JarFile;
import java.util.regex.Matcher;
import java.util.zip.ZipEntry;

import org.ow2.util.archive.api.IArchive;
import org.ow2.util.archive.impl.ArchiveManager;
import org.ow2.util.archive.impl.DirectoryArchiveImpl;
import org.ow2.util.ee.deploy.api.deployable.EARDeployable;
import org.ow2.util.ee.deploy.api.deployable.IDeployable;
import org.ow2.util.ee.deploy.api.deployable.WARDeployable;
import org.ow2.util.ee.deploy.api.helper.DeployableHelperException;
import org.ow2.util.ee.deploy.impl.deployable.EARDeployableImpl;
import org.ow2.util.ee.deploy.impl.helper.DeployableEntry;
import org.ow2.util.ee.deploy.impl.helper.DeployableHelper;
import org.ow2.util.ee.deploy.impl.helper.ModifyDeployableHelper;
import org.ow2.util.ee.deploy.impl.helper.PackDeployableHelper;
import org.ow2.util.file.FileUtils;
import org.testng.Assert;
import org.testng.annotations.AfterClass;
import org.testng.annotations.BeforeClass;
import org.testng.annotations.Test;

/**
 * Test the helper classes of Deployable.
 * @author Florent BENOIT
 */
public class TestDeployable {

    /**
     * Temp folder.
     */
    private File tmpFolder = null;

    /**
     * Temp unpack folder.
     */
    private File tmpUnpackFolder = null;

    /**
     * Temp pack folder.
     */
    private File tmpPackFolder = null;

    /**
     * Ear Deployable.
     */
    private EARDeployable earDeployable;

    /**
     * War Deployable.
     */
    private WARDeployable warDeployable;

    /**
     * Create directories for the test.
     * @throws IOException if directories are not created
     * @throws DeployableHelperException if deployables are not created
     */
    @BeforeClass
    public void init() throws IOException, DeployableHelperException {
        // Build a temp directory (which is root of all unpack)
        tmpFolder = new File(System.getProperty("java.io.tmpdir") + File.separator + System.getProperty("user.name", "default")
                + "testNG-deployable");


        tmpFolder.mkdirs();
        // Delete first (if not cleanly deleted)
        remove();
        tmpFolder.mkdirs();

        // Create folder for unpacking
        tmpUnpackFolder = new File(tmpFolder, "unpack");
        tmpUnpackFolder.mkdirs();

        // Create pack folder
        tmpPackFolder = new File(tmpFolder, "pack");
        tmpPackFolder.mkdirs();

        // Then recreate
        createArchives();
    }

    /**
     * Creates deployable.
     * @throws IOException if IO problems
     * @throws DeployableHelperException if the Deployable are not created.
     */
    public void createArchives() throws IOException, DeployableHelperException {
        createWarDeployable();
        createEarDeployable();
    }

    /***
     * Create a war file with dummy files inside.
     * @throws DeployableHelperException if the deployable is not created.
     * @throws IOException if war is not created
     */
    public void createWarDeployable() throws IOException, DeployableHelperException {
        // Create file (a directory)
        File warFolder = new File(tmpFolder, "mytest.war");
        warFolder.mkdir();

        // Create files inside this war
        File tmpFile = new File(warFolder, "MyDummy.file");
        tmpFile.getParentFile().mkdirs();

        FileWriter fWriter = new FileWriter(tmpFile);
        fWriter.write("first entry");
        fWriter.flush();
        fWriter.close();

        // 2nd entry
        tmpFile = new File(warFolder, "myDirectory" + File.separator + "MyDummy2.file");
        tmpFile.getParentFile().mkdirs();
        fWriter = new FileWriter(tmpFile);
        fWriter.write("second entry");
        fWriter.flush();
        fWriter.close();

        // 2nd entry
        tmpFile = new File(warFolder, "otherDirectory" + File.separator + "subDirectory" + File.separator + "MyDummy3.file");
        tmpFile.getParentFile().mkdirs();
        fWriter = new FileWriter(tmpFile);
        fWriter.write("second entry");
        fWriter.flush();
        fWriter.close();

        // Build Archive
        IArchive archive = ArchiveManager.getInstance().getArchive(warFolder);

        // Assume it is a Directory archive
        Assert.assertTrue(DirectoryArchiveImpl.class.isInstance(archive));

        // Build deployable
        IDeployable<?> deployable = DeployableHelper.getDeployable(archive);

        // Check
        Assert.assertTrue(WARDeployable.class.isInstance(deployable));

        warDeployable = WARDeployable.class.cast(deployable);

    }

    /***
     * Create an EAR file with dummy files inside.
     * @throws DeployableHelperException if the deployable is not created.
     * @throws IOException if ear is not created
     */
    public void createEarDeployable() throws IOException, DeployableHelperException {
        // Create file (a directory)
        File earFolder = new File(tmpFolder, "mytest.ear");
        earFolder.mkdir();


        // Build Archive
        IArchive archive = ArchiveManager.getInstance().getArchive(earFolder);

        // Assume it is a Directory archive
        Assert.assertTrue(DirectoryArchiveImpl.class.isInstance(archive));

        // Build deployable
        IDeployable<?> deployable = DeployableHelper.getDeployable(archive);

        // Check
        Assert.assertTrue(EARDeployable.class.isInstance(deployable));

        earDeployable = EARDeployable.class.cast(deployable);

        ((EARDeployableImpl) earDeployable).addDeployable(warDeployable);

        // Check war is added
        List<WARDeployable> warList = earDeployable.getWARDeployables();
        Assert.assertEquals(warList.size(), 1);

        // Right content ?
        Assert.assertEquals(warDeployable, warList.get(0));


    }

    /**
     * Remote the temp folder created at the end of the test.
     */
    @AfterClass
    public void remove() {
       // Delete temp folder
       Assert.assertTrue(FileUtils.delete(tmpFolder));
    }


    /**
     * Test if an entry can be added to a deployable.
     * @throws Exception if there is a failure
     */
    @Test
    public void testAddFile() throws Exception {
        // Create file with a content
        File tmpFile = new File(tmpFolder, "myEntryToAdd");
        Writer fWriter = new FileWriter(tmpFile);
        fWriter.write("entry-added");
        fWriter.flush();
        fWriter.close();

        DeployableEntry deployableEntry = new DeployableEntry();
        deployableEntry.setFile(tmpFile);

        deployableEntry.setName("my/entry/to/add".replaceAll("/", Matcher.quoteReplacement(File.separator)));

        List<DeployableEntry> deployablesEntries = new ArrayList<DeployableEntry>();
        deployablesEntries.add(deployableEntry);

        // build an entry to add in the given war deployable
        ModifyDeployableHelper.addInDeployable(warDeployable, deployablesEntries, tmpUnpackFolder);

    }

    /**
     * Test if the entry has been added in the deployable at the right place.
     * @throws Exception if there is a failure
     */
    @Test(dependsOnMethods = "testAddFile")
    public void testModifiedDeployable() throws Exception {
        WARDeployable unpackedDeployable = warDeployable.getUnpackedDeployable();
        Assert.assertNotNull(unpackedDeployable);

        // Entry has been added ?
        URL entryAdded = unpackedDeployable.getArchive().getResource("my/entry/to/add");
        Assert.assertNotNull(entryAdded);

        // Check content
        URLConnection urlConnection = entryAdded.openConnection();
        urlConnection.setUseCaches(false);
        InputStream is = urlConnection.getInputStream();
        InputStreamReader inputStreamReader = new InputStreamReader(is);
        BufferedReader bufferedReader = new BufferedReader(inputStreamReader);
        String line = bufferedReader.readLine();
        Assert.assertEquals(line, "entry-added");
        bufferedReader.close();
        inputStreamReader.close();
        is.close();
    }

    /**
     * Test if shorter name are valid.
     * @throws Exception
     */
    @Test
    public void testShortName() {
        Assert.assertEquals(warDeployable.getShortName(), "mytest.war");
        Assert.assertEquals(earDeployable.getShortName(), "mytest.ear");
    }

    /**
     * Test if module name are valid.
     */
    @Test
    public void testModuleName() {
        Assert.assertEquals(warDeployable.getModuleName(), "mytest");
        Assert.assertEquals(earDeployable.getModuleName(), "mytest");
    }

    /**
     * Test if the archive can be packed.
     * @throws Exception if there is a failure
     */
    @Test(dependsOnMethods = "testModifiedDeployable")
    public void testPackDeployable() throws Exception {
        File packedWar = PackDeployableHelper.pack(warDeployable, tmpPackFolder, "mypacked.war");

        // Build Jar File on it
        JarFile jarFile = new JarFile(packedWar);

        // Get entry
        ZipEntry jarEntry = jarFile.getEntry("my/entry/to/add");
        Assert.assertNotNull(jarEntry);

        // Read Stream
        InputStream is = jarFile.getInputStream(jarEntry);
        Assert.assertNotNull(is);

        // Verify
        InputStreamReader inputStreamReader = new InputStreamReader(is);
        BufferedReader bufferedReader = new BufferedReader(inputStreamReader);
        String line = bufferedReader.readLine();
        Assert.assertEquals(line, "entry-added");
        bufferedReader.close();
        inputStreamReader.close();
        is.close();
        jarFile.close();

    }

}
